﻿/******************************  Переменные перечисления  **********************************/
/// <summary>
/// Перечисление - Тип протокола передачи данных
/// </summary>
enum ConnectionTypeEnum
{
    /// <summary>
    /// Не выбран
    /// </summary>
    CT_None = 0,   
    /// <summary>
    /// Serial порт
    /// </summary>
    CT_Serial = 1,
    /// <summary>
    /// UDP
    /// </summary>	
    CT_UDP = 2
};

/// <summary>
/// Переменная определяющая тип связи с компьютером
/// </summary>
ConnectionTypeEnum connectionType = CT_Serial; 

/// <summary>
/// Перечисление - Тип операции 
/// </summary>
enum OperationTypeEnum     
{
    /// <summary>
    /// Не выбрана
    /// </summary>
    OT_None = 0,   
    /// <summary>
    /// Отправка данных с Arduino
    /// </summary>
    OT_SendDataFromArduino = 1,
    /// <summary>
    /// Отправка данных с PC
    /// </summary>
    OT_SendDataFromPC = 2,
    /// <summary>
    /// Получение данных с Arduino
    /// </summary>
    OT_ReciveDataFromArduino  = 3,
    /// <summary>
    /// Получение данных с PC
    /// </summary>
    OT_ReciveDataFromPC = 4,
    /// <summary> 
    /// Проверка вернувшихся данных с Arduino
    /// </summary>
    OT_CheckReturnFromArduino = 5,
    /// <summary>
    /// Проверка вернувшихся данных с PC
    /// </summary>
    OT_CheckReturnFromPC = 6,
    /// <summary>
    /// Старт ТПА
    /// </summary>
    OT_StartTPA = 7,
    /// <summary>
    /// Начало инициализации Arduino
    /// </summary>
    OT_InitArduinoBegin = 8,
    /// <summary>
    /// Инициализация Arduino
    /// </summary>
    OT_InitArduino = 9,
    /// <summary>
    /// Конец инициализации Arduino
    /// </summary>
    OT_InitArduinoFinish = 10,
    /// <summary>
    /// Начало настройки Arduino
    /// </summary>
    OT_SetupArduinoBegin = 11,
    /// <summary>
    /// Настройка Arduino
    /// </summary>
    OT_SetupArduino = 12,
    /// <summary>
    /// Конец настройки Arduino
    /// </summary>
    OT_SetupArduinoFinish = 13,
    /// <summary>
    /// Проверка соединения с Arduino
    /// </summary>
    OT_CheckConnectionFromArduino = 14,
    /// <summary>
    /// Проверка соединения с PC
    /// </summary>
    OT_CheckConnectionFromPC = 15,
	/// <summary>
	/// Команда ручного режима 
	/// </summary>
	OT_CommandManualMode = 16,
	/// <summary>
    /// Стоп ТПА
    /// </summary>
    OT_StopTPA = 17
};

/// <summary>
/// Переменная хранящая тип операции
/// </summary>
OperationTypeEnum operationType = OT_None; 

/// <summary>
/// Перечисление - Тип параметра
/// </summary>
enum ParametrEnum
{
    /// <summary>
    /// Не установлен
    /// </summary>
    P_None = 0
};

/// <summary>
/// Перечисление - Тип  данных параметра
/// </summary>
enum DataTypeEnum 
{
    /// <summary>
    /// Не указан
    /// </summary>
    DT_None = 0,
    /// <summary>
    /// Int
    /// </summary>
    DT_Int = 1
};
/******************************  /Переменные перечисления  **********************************/

/******************************  Переменные для работы с Serial портом  **********************************/
/// <summary>
/// Константа - Скорость Serial порта
/// </summary>
#define speedSerial 115200    
/******************************  /Переменные для работы с Serial портом **********************************/

/******************************  Переменные для работы передачи данных  **********************************/
/// <summary>
/// Константа - Символ новой строки
/// </summary>
#define newLine '\n'
/// <summary>
/// Константа - Символ начала пакета
/// </summary>
#define beginChar '?'
/// <summary>
/// Константа - Символ разделитель параметров пакета
/// </summary>
#define breakChar ';'
/// <summary>
/// Константа - Символ конца пакета
/// </summary>
#define endChar '!'
/// <summary>
/// Константа - Символ для очистки значения char
/// </summary>
#define clearChar '\0'

/// <summary>
/// Переменная хранящая последний полученный char
/// </summary>
char incomingChar;

/// <summary>
/// Переменная хранящая индекс текущего char в массиве хранящего пакет данных
/// </summary>
int indexCharInArray;

/// <summary>
/// Переменная хранящая завершено ли получения данных
/// </summary>
boolean finishReceiveData = false;

/// <summary>
/// Константа хранящая мак. размер массива хранящего пакет данных
/// </summary>
#define receiveSizeArrayCOM 33
/// <summary>
/// Массив хранящий пакет данных
/// </summary>
char receiveCharArrayCOM[receiveSizeArrayCOM];

/// <summary>
/// Переменная хранящее номер пакета в системе
/// </summary>
int globalNumberPacket = 0;

/// <summary>
/// Переменная хранящее номер символа начала пакета в пакете
/// </summary>
int indexBeginCharInPackage;
/// <summary>
/// Переменная хранящее номер символа конца пакета в пакете
/// </summary>
int indexEndCharInPackage;
/// <summary>
/// Переменная кол-во символов разделителя блоков данных в пакете
/// </summary>
int breakCharCountInPackage;

/// <summary>
/// Переменная хранящее номер пакета
/// </summary>
String numberPacket;
/// <summary>
/// Переменная хранящее длину пакета
/// </summary>
String lenPacket; 
/// <summary>
/// Переменная хранящее тип операции пакета
/// </summary>
String operationTypePacket;
/// <summary>
/// Переменная хранящее тип параметра пакета
/// </summary>
String parametrPacket;
/// <summary>
/// Переменная хранящее тип данных параметра пакета
/// </summary>
String dataTypePacket;
/// <summary>
/// Переменная хранящее значение параметра пакета
/// </summary>
String valuePacket;

/// <summary>
/// Переменная хранящее номер пакета
/// </summary>
int receiveNumberIntPacket = 0;
/// <summary>
/// Переменная хранящее тип операции пакета
/// </summary>
int receiveLenIntPacket;
/// <summary>
/// Переменная хранящее тип операции пакета
/// </summary>
int receiveOperationTypeIntPacket;

/// <summary>
/// Переменная хранящее реальную длину пакета, проверка после получения
/// </summary>
int receiveLength;
/******************************  /Переменные для работы передачи данных  **********************************/

/******************************  Переменные запуска ТПА  **********************************/
/// <summary>
/// Переменная хранящее состояние программы запущена или нет
/// </summary>
boolean beginWorkTPA = false; 

/// <summary>
/// Переменная хранящее состояние программы инициализирован ли ТПА
/// </summary>
boolean beginInitializationTPA = false;
/******************************  /Переменные запуска ТПА  **********************************/

/******************************  Переменные отладки  **********************************/
/// <summary>
/// Переменная хранящее время старта цикла, для проверки скорости выполнения цикла
/// </summary>
unsigned long timeCycleWork = 0;

#define IS_DEBUG true
/******************************  /Переменные отладки  **********************************/

/// <summary>
/// Настройка программы
/// </summary>
void setup ()
{
	// TODO: Добавить описание
	switch (connectionType) 
	{
		case CT_Serial:
			Serial.begin(speedSerial);
			break;
	}	
}

/// <summary>
/// Основной цикл
/// </summary>
void loop()
{   
	// Проверка времени обработки данных
	//timeCycleWork = millis();
	// /Проверка времени обработки данных

	// Очистка переменных получения данных, после получения данных
	ClearReceiveDataVariable();

	// Запущена или программа ТПА
	if (!beginWorkTPA)
	{
		// Программа ТПА не запущена

		// Выбираем протокол передачи данных
		switch (connectionType) 
		{
			case CT_Serial:
				// Протокол передачи данных на основе Serial порта

				// Получение данных с Serial порта
				GetReceiveDataFromSerial();
				break;
		}	

		// Парсинг полученных данных
		ParseReceiveData();

		// Обрываем выполнение loop
		return;
	}
 
	// Состояние программы инициализирован ли ТПА
	if (!beginInitializationTPA)
	{	
		// ТПА не инициализирован
	
		// Устанавливаем состояние программы инициализирован
		beginInitializationTPA = true;
	}

	// Выбираем протокол передачи данных
	switch (connectionType) 
	{
		case CT_Serial:
			// Протокол передачи данных на основе Serial порта

			// Получение данных с Serial порта
			GetReceiveDataFromSerial();
			break;
	}	
    
	// Парсинг полученных данных
	ParseReceiveData();

	delay(3000);
	SendDataFromSerial("Hello", OT_SendDataFromArduino, P_None, DT_Int);	        

	// Проверка времени обработки данных
	//unsigned long workTime = millis() - timeCycleWork;
	//Serial.println(workTime);
	// /Проверка времени обработки данных
}

/******************************  Функции работы с Serial портом  **********************************/
/// <summary>
/// Получение данных с Serial порта
/// </summary>
void GetReceiveDataFromSerial()
{
	// TODO: Добавить описание 
	while(Serial.available())
	{
		incomingChar = (char)Serial.read();
		if(incomingChar != newLine)
		{
			receiveCharArrayCOM[indexCharInArray++] = incomingChar;
		}
		else
		{
			finishReceiveData = true;
		}
	  
		delayMicroseconds(100); 
	}
}

/// <summary>
/// Отправка данных по Serial порту
/// </summary>
void SendDataFromSerial(String SendValue, int OperationType, int Parametr, int DataType)
{
	// TODO: Добавить описание 
	/*if (globalNumberPacket > 500)
	{
		Serial.println("Errore before");
	}
	globalNumberPacket++;
	if (globalNumberPacket > 500)
	{
		Serial.println("Errore after");
	}*/
	RepeatSendDataFromSerial(SendValue, OperationType, Parametr, DataType);          
}

/// <summary>
/// TODO: Добавить описание 
/// </summary>
void SendConfirmationDataFromSerial(String SendValue, int Parametr, int DataType)
{
	// TODO: Добавить описание 
	RepeatSendDataFromSerial(SendValue, OT_CheckReturnFromArduino, Parametr, DataType);          
}

/// <summary>
/// Повтор отправки данных по Serial порту
/// </summary>
void RepeatSendDataFromSerial(String SendValue, int OperationType, int Parametr, int DataType)
{
	// TODO: Добавить описание 
	int sendValueLen = String(globalNumberPacket).length() +  String(OperationType).length() + String(Parametr).length() + String(DataType).length() + SendValue.length() + 7;
	sendValueLen += String(sendValueLen).length();
                  
	String sendPacket = beginChar + String(globalNumberPacket) + breakChar + String(sendValueLen) + breakChar + OperationType + breakChar + Parametr + breakChar + DataType + breakChar + SendValue + endChar;     
                                    
	Serial.println(sendPacket);
}
/******************************  /Функции работы с Serial портом  **********************************/

/******************************  Функции работы с полученными данными  **********************************/
/// <summary>
/// Очистка переменных получения данных, после получения данный
/// </summary>
void ClearReceiveDataVariable()
{
	// TODO: Добавить описание 
	indexCharInArray = 0;
	memset(receiveCharArrayCOM,clearChar,receiveSizeArrayCOM); 

	indexBeginCharInPackage = indexEndCharInPackage = breakCharCountInPackage = receiveLenIntPacket = receiveLength = receiveOperationTypeIntPacket = 0;
	numberPacket = lenPacket = operationTypePacket = parametrPacket = dataTypePacket = valuePacket = "";
		
	finishReceiveData = false;
}

/// <summary>
/// Парсинг полученных данных
/// </summary>
void ParseReceiveData()
{
	// Завершено ли получения данных
	if (finishReceiveData)
	{
		// Получаем длину пакета
		receiveLength = strlen(receiveCharArrayCOM);
		
		// Пробегаемся по массиву хранящий пакет данных
		for (int i = 0; i < receiveLength; i++)
		{
			switch (receiveCharArrayCOM[i]) {
				case beginChar:
					// Получаем индекс символ начала пакета в пакете
					indexBeginCharInPackage = i;
					break;
				case endChar:
					// Получаем индекс символ конца пакета в пакете
					indexEndCharInPackage = i;
					break;
				case breakChar:
					// Получаем кол-во символов разделителя блоков данных в пакете
					breakCharCountInPackage++;
					break;
				default:
					// В зависимости от номера блока данных в пакете, заполняем переменные блоков
					switch (breakCharCountInPackage) {
						case 0:
							// Номер пакета в системе
							numberPacket += receiveCharArrayCOM[i];
							break;
						case 1:
							// Длина пакета
							lenPacket += receiveCharArrayCOM[i];
							break;
						case 2:
							// Тип операции пакета
							operationTypePacket += receiveCharArrayCOM[i];
							break;
						case 3:
							// Тип операции пакета
							parametrPacket += receiveCharArrayCOM[i];
							break;
						case 4:
							// Тип данных параметра пакета
							dataTypePacket += receiveCharArrayCOM[i];
							break;
						case 5:
							// Значение параметра пакета
							valuePacket += receiveCharArrayCOM[i];					
							break;
					} 
					break;
			}
		}
       
		// Номер пакета в системе, Int
		receiveNumberIntPacket = numberPacket.toInt();

		// Номер пакета в системе, Int
		receiveLenIntPacket = lenPacket.toInt();

		// Тип операции пакета, Int
		receiveOperationTypeIntPacket = operationTypePacket.toInt();

		// Проверка корректности полученного пакета 
		if (indexBeginCharInPackage == 0 && breakCharCountInPackage == 5	// Индекс символа начала пакета, должен быть равен 0, кол-во блоков данных должно быть 5	
				&& indexEndCharInPackage == (receiveLength - 1)				// Индекс символа конца пакета, должен быть равен значению блока длины пакета - 1
				&& receiveLenIntPacket == receiveLength						// Реальная длина пакета должна быть равна значению блока длины пакета
				)
		{
			// Определяем тип операции пакета
			switch(receiveOperationTypeIntPacket)
			{
				case OT_StartTPA:
					// Старт ТПА

					SendConfirmationDataFromSerial(valuePacket, parametrPacket.toInt(), dataTypePacket.toInt());

					// Устанавливаем состояние программы запущена
					beginWorkTPA = true;
					break;	
				default:
					int globalNumberPacketAndOne = globalNumberPacket;//globalNumberPacket + 1;
					if (receiveOperationTypeIntPacket != OT_CheckReturnFromPC && receiveNumberIntPacket == globalNumberPacketAndOne)
					{			
						SendConfirmationDataFromSerial(valuePacket, parametrPacket.toInt(), dataTypePacket.toInt());
					}
					else 
					{
						if (receiveNumberIntPacket == globalNumberPacketAndOne)
						{
							switch(receiveOperationTypeIntPacket)
							{
								case OT_StopTPA:
									// Стоп ТПА
							
									SendConfirmationDataFromSerial(valuePacket, parametrPacket.toInt(), dataTypePacket.toInt());
								
									//globalNumberPacket = 0;

									// Устанавливаем состояние программы остановлена
									beginWorkTPA = false;
									break;
							}
						}
						else 
						{
							ErroreDataReceived("globalNumberPacket",0);
						}
					}		
					break;
			}
		}
		else 
		{
			ErroreDataReceived("",0);
		}              								
	}
}
/******************************  /Функции работы с полученными данными  **********************************/

/******************************  Функции работы с ошибками  **********************************/
/// <summary>
/// Ошибка получения данных
/// </summary>
void ErroreDataReceived(String DataReceived, int TypeErrore)
{
	// TODO: Добавить описание 
	Errore(DataReceived, 0);
}

/// <summary>
/// Ошибка получения данных
/// </summary>
void Errore(String Data, int TypeErrore)
{
	// TODO: Добавить описание 
	//FullStopTPA();

	Log(Data, TypeErrore, true);
}
/******************************  /Функции работы с ошибками  **********************************/

/******************************  Функции работы с логом  **********************************/
/// <summary>
/// Логирование
/// </summary>
void Log(String Data, int TypeLog, bool isErrore)
{
	// TODO: Добавить описание 
	switch (connectionType) 
	{
		case CT_Serial:
			Serial.println(Data);
			break;
	}
}
/******************************  /Функции работы с логом  **********************************/

